#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>

#include <ihx.h>

#include "ppdrv.h"
#include "dummydrv.h"
#include "ftdidrv.h"
#include "chip.h"

#define DEFAULT_CHIP	"27c040"

static const char DEFAULT_DRIVER[] = "ftdi";

static bool quiet = false;

static void usage(const char *exe) {
	printf("Usage: %s [options] -t chiptype file\n", exe);
	puts("Options:");
	puts(" -b - binary file (default is Intel hex)");
	puts(" -d <driver> - Available drivers: ftdi, pp, dummy");
	puts(" -q - quiet");
	puts(" -h - this help");
	puts("chiptype examples (same type): a15 27c512 512");
}

static TROM * AllocDriver(const char *name) {
	if (!strcmp(name, "ftdi"))
		return new FtdiDrv();
	if (!strcmp(name, "dummy"))
		return new DummyDrv();
	if (!strcmp(name, "pp")) {
		PPDrv *ppd = new PPDrv();
		if (!ppd->Open(PPDrv::PP_address[0]))
			return NULL;
		return ppd;
	}

	warnx("Unknown driver!");
	return NULL;
}

static void fb(unsigned long sent, unsigned long size, void *arg) {
	if (sent == size) {
		fprintf(stderr, "Done!\n");
		return;
	}
	int posto = (int)(sent * 100 / size);
	if (*((int*)arg) != posto) {
		fprintf(stderr, "%d%%\r", posto);
		fflush(stderr);
		*((int*)arg) = posto;
	}
}

static bool SendToTROM(const Chip &chip, TROM &trom, const ihx::Buffer &buf, unsigned int offset) {
	unsigned long sz = buf.LastAddress() + 1;
	int oldp = 100;
	void *b;

	if (sz > chip.Size()) sz = chip.Size();

	if (offset > chip.Size()) {
		warnx("Offset greater than chip size.");
		return false;
	}

	if (!quiet) {
		printf("Sending 0x%05X (%d) bytes", (unsigned int)sz, (unsigned int)sz);
		if (offset)
			printf(" from 0x%05X", offset);
		printf("...\n");
	}
	b = buf.GetBINBlock(sz);
	trom.SendData(b, sz, chip.GetTROMAddress() + offset, quiet ? NULL : fb, &oldp);
	buf.FreeBINBlock(b);
	return true;
}

int main(int argn, char *argv[]) {
	TROM *trom;
	const char *drvname = DEFAULT_DRIVER;
	int c, retval = 0;
	bool defaultchip = true, binaryfile = false;
	Chip chip(DEFAULT_CHIP);
	const char *filename;
	unsigned int offset = 0;
	ihx::Buffer buf;

	while ((c = getopt(argn, argv, "d:t:qhbo:")) != -1) {
	switch (c) {
		case 'd':
			drvname = optarg;
			break;
		case 't':
			if (!chip.ConfigureFrom(optarg))
				return false;
			defaultchip = false;
			break;
		case 'o': {
			char *ep;
			offset = strtoul(optarg, &ep, 0);
			if (ep == optarg) {
				warnx("Invalid offset");
				return 1;
			}
			break;
			}
		case 'q':
			quiet = true;
			break;
		case 'b':
			binaryfile = true;
			break;
		case 'h':
			usage(*argv);
			return 0;
		case '?':
			usage(*argv);
			return 1;
	}
	}

	if (optind + 1 != argn) {
		usage(*argv);
		return 1;
	}
	filename = argv[optind];

	if (!quiet) {
		const bool *js = chip.GetJumperSettings();
		printf("Jumpers: %d%d%d%d\n", js[0], js[1], js[2], js[3]);
	}

	if (defaultchip) {
		warnx("Warning: Using default chip " DEFAULT_CHIP "!");
	}

	// učita bin ili ihx
	if (ihx::Buffer::Error e = binaryfile ? buf.ReadBIN(filename) : buf.ReadIHX(filename)) {
		warnx("Error while loading file: %s\n", buf.GetErrorString(e));
		return 1;
	}

	// učita i inicijalizira drajvere
	if (!quiet) {
		printf("Loading driver \"%s\"...\n", drvname);
	}
	trom = AllocDriver(drvname);
	if (!trom) return 1;

	if (!trom->Init()) {
		retval = 1;
		goto kraj;
	}

	if (!trom->Test()) {
		fprintf(stderr, "TurboROM not detected!\n");
		retval = 1;
		goto kraj;
	}

	if (!SendToTROM(chip, *trom, buf, offset))
		retval = 1;

kraj:
	trom->Shutdown();
	delete trom;
	return retval;
}
